// fonted.cpp		ver.0.02
//copyright 2005-2012 Beer[Tsutsumi Hiroyuki]

#include	<Be.h>
#if	B_BEOS_VERSION>=0x0520
#include	<Locale.h>
typedef	const char	Char;
#else
#define	_T(str)	str
typedef	char		Char;
#endif

typedef	signed short	sint16;

#define	MAX_COMPONENT	8

#define	SELECTVIEWCOLOR	0x00ffcc44
#define	TOOLVIEWCOLOR	0x00ffcc44

enum{	B_QUATERNARY_MOUSE_BUTTON=0x08	};		//4th-button

//for input(tablet,mouse,key)
enum{				//flags
	TABLET_XY=		0x0001,		//be:tablet_x,be:tablet_y
	TABLET_TILT_XY=	0x0002,		//be:tablet_tilt_x,be:tablet_tilt_y
	TABLET_PRESSURE=0x0004,		//be:tablet_pressure
	TABLET_ERASER=	0x0008,		//be:tablet_eraser
	BUTTONS=		0x0010,		//buttons
	MODIFIERS=		0x0020,		//modifiers
	WHEN=			0x0100,		//when
	WHERE=			0x0200,		//be:view_where
	TRANSIT=		0x0400,		//be:transit
};

enum{				//mouse
	NONE=0,
	PRESS,
	RELEASE,
	PRESSMOVE,
	RELEASEMOVE,
	CTRLMOVE,
	CANSEL
};

enum{				//object-type
	PICK=0,
	FREE=0x10,
	RECT=0x30,
	ELLIPSE=0x40,
	POLY=0x50,
	ROUND,
};

typedef struct{				//for input
	uint16	flags;
	float	x,y,tilt_x,tilt_y,pressure;
	uint32	eraser,buttons,modifiers,transit;
	uint64	when;
	BPoint	where;
}input;

typedef struct{
	uchar	f;				//0.frame-off  1.frame-on
	BRect	rect;
}s_sel;

typedef struct{		//for init ttf
	bool	ub[128];	//use unicode_block
	bool	c[0x10000];	//use unicode
	int		chrs;
	int		glfs;
}inittf;
//========================================================
class shape
{
	public:
		shape(int c);
		~shape();
		shape*	copy();
		bool	get(const int &no,BPoint &p,char &f);
		void	set(const int &no,const BPoint &p,const char &f);
		int		select_chk(const BPoint &p,float pz);	//point size
		int		select_chk(const BRect &r);
		void	move(const BPoint &d);
		void	inspoint(const BPoint &d);
		bool	delpoint();
		void	tglpoint();
		void	reverse();
		void	draw(BView *v,uint32 col=0);
		void	drawframe(BView *v,float pz);	//point size
		void	ctrlp2form();
		void	trans(const BRect &from,const BRect &to);		//for canvas
		shape*	transcopy(int16 ox,int16 oy,int16 sc00,int16 sc01,int16 sc10,int16 sc11);
		bool	ispositive();
		int		sel_pt;		//-1.inner  0.notselect  1-.point(corner)
		bool	addf;
		BRect	rec;
//	private:
		int		pts;			//count of points
		BPoint	*pt;			//points
		char	*pf;			//0.anchor_point  1.control_point
};

class glyph
{
	public:
		glyph(uint16 no=0xffff);
		~glyph();
		void	add_shape(shape *s);
		void	del_sel();
		void	tgl_sel();
		void	rev_sel();
		void	copy_sel(const BPoint &d);
		void	move_sel(const BPoint &d);
		void	copy();
		void	paste();
		void	sel_check();
		uchar	sel_check(const BPoint &p,float pz);	//point size
		bool	sel_check(const BRect &r);
		void	sel_none();
		void	sel_to_top();
		void	sel_to_bot();
		void	sel_first();
		void	sel_last();
		void	sel_next();
		void	sel_back();
		void	trans(const BRect &from,const BRect &to);
		void	draw(BView *v,uint32 col=0);
		void	drawframe(BView *v,float pz);	//point size
		int		del_compo(uchar no);
		int		add_compo(glyph *sub,int16 x=0,int16 y=0,uint16 k00=0x4000,uint16 k01=0x0000,uint16 k10=0x0000,uint16 k11=0x4000);
		void	clr_composhape();			//shape for subglyph_bounds
		void	create_composhape();		//shape for subglyph_bounds
		BRect	bounds()const;
		int		dep;			//0.parent  +.child
		int		cnt;			//shape count
		shape	**sh;
		int		gc;				//0.simple  +.glyf count
		glyph	**gl;
		int16	*ox,*oy;					//glyf offset
		uint16	*sc00,*sc01,*sc10,*sc11;	//scale
		uint16	gno;	//glyf.no
};

class object
{
	public:
		object(int points=1);
		~object();
		virtual int		create(const uchar &mf,const BPoint &mp,uchar psize);	//mf=1 down,mf=2 up,mf=3 moved
		virtual void	drawframe(BView *v,float scale);
		virtual shape**	to_shape();
		uchar	objtype;
		int		sel_pt;		//-1.inner  0.notselect  1-.point(corner)
		int		cpts;		//created points	-1.complete
		int		pts;
		BPoint	*pt;
};

class object_r:public object
{
	public:
		object_r();
		int		create(const uchar &mf,const BPoint &mp,uchar psize);
		virtual void	drawframe(BView *v,float scale);
};

class object_l:public object
{
	public:
		object_l();
		int		create(const uchar &mf,const BPoint &mp,uchar psize);
};

class obj_free:public object
{
	public:
		obj_free();
		int		create(const uchar &mf,const BPoint &mp,uchar psize);
		void	drawframe(BView *v,float scale);
		shape**	to_shape();
		uchar	psz;	//pensize
};

class obj_rect:public object_r
{
	public:
		obj_rect();
		shape**	to_shape();
};

class obj_ellipse:public object_r
{
	public:
		obj_ellipse();
		void	drawframe(BView *v,float scale);
		shape**	to_shape();
};

class obj_polygon:public object_l
{
	public:
		obj_polygon();
		void	drawframe(BView *v,float scale);
		shape**	to_shape();
};

class obj_round:public object_l
{
	public:
		obj_round();
		void	drawframe(BView *v,float scale);
		shape**	to_shape();
};
//========================================================
class bitmap
{
	public:
		bitmap(uint16 no=0xffff,uchar width=8,uchar height=8,uchar depth=1);
		~bitmap();
		void	setvrect(const BRect &to);
		void	draw(BView *v);
		int		create(const uchar &mf,const BPoint &mp);
	public:
		uint16	gno;
		uchar	ppemx,ppemy;
		uchar	dep;	//bitmap-depth
		uchar	w,h;	//datasize
		char	bx,by;	//bearing
		uchar	ad;		//advance
		uchar	*dat;
	public:
		BRect	vr;
};
//========================================================
class	ttf;
//class	tablecontrol;
class	glyftoolview;
class	glyfeditview;
class	ebdttoolview;
class	ebdteditview;

class ttf_table:public BView
{
	public:
		ttf_table(ttf *p,uint32 tname,uchar *dat=NULL,uint32 size=0);	//NULL...create default-table
		~ttf_table();
		virtual void	show(bool f);
		uint32	sumB32();
		ttf		*pak;
		uint32	tbl_name;
		uchar	*tbl_dat;
		int		tbl_size;
	private:
};

class ttf_os2:public ttf_table
{
	public:
		ttf_os2(ttf *p,uchar *dat=NULL,uint32 size=0);
		~ttf_os2();
		void	Draw(BRect rect);
		void	redraw();
		void	create_default();
	private:
};

class ttf_cmap:public ttf_table
{
	public:
		ttf_cmap(ttf *p,uchar *dat=NULL,uint32 size=0);
		~ttf_cmap();
		void	AttachedToWindow();
		void	Draw(BRect rect);
		void	MessageReceived(BMessage *msg);
		void	redraw();
		void	create_default();
		int		get_tbl();		//cmap > c2g	0.ok  -1.not_found  +.err
		int		set_tbl();		//c2g > cmap
		uint16	get_gno(uint16 chr);
		uint16	get_chr(uint16 gno);
		int		set_chr(uint16 chr_st,uint16 chr_ed,uint16 gno_st);
	private:
		BTextControl	*tc_st,*tc_ed,*tc_gno;
		BButton			*pb_set;
	private:
		uint16	*c2g;
};

class ttf_glyf:public ttf_table
{
	public:
		ttf_glyf(ttf *p,uchar *dat=NULL,uint32 size=0);
		~ttf_glyf();
		int		decode(uint32 off,uint32 len,glyph *gl);	//0.ok  -1.not_found  +.err
		int		encode(const glyph *gl,uint32 *off,uint32 *len);	///del&remake dat
		int		get_info(uint16 g_no,uint16 *dep,uint16 *els,uint16 *cts,uint16 *pts);
		int		get_rect(uint16 g_no,int16 *xmin,int16 *ymin,int16 *xmax,int16 *ymax);
		glyftoolview	*tv;
		glyfeditview	*ev;
	private:
};

class ttf_head:public ttf_table
{
	public:
		ttf_head(ttf *p,uchar *dat=NULL,uint32 size=0);
		~ttf_head();
		void	AttachedToWindow();
		void	Draw(BRect rect);
		void	MessageReceived(BMessage *msg);
		void	redraw();
		void	create_default();
		BRect	bounds();
		bool	is32loca();
		uint16	unitsperem();
		void	update();		//set time
		void	setsum(uint32 sum=0xb1b0afba);
	private:
		BTextControl	*tc1;
	private:
		time_t	t1904;	//1904/01/01_00:00:00
};

class ttf_hhea:public ttf_table
{
	public:
		ttf_hhea(ttf *p,uchar *dat=NULL,uint32 size=0);
		~ttf_hhea();
		void	Draw(BRect rect);
		void	redraw();
		void	create_default();
		uint16	numhmtx();
	private:
};

class ttf_hmtx:public ttf_table
{
	public:
		ttf_hmtx(ttf *p,uchar *dat=NULL,uint32 size=0);
		~ttf_hmtx();
		void	AttachedToWindow();
		void	Draw(BRect rect);
		void	MessageReceived(BMessage *msg);
//		void	redraw();
		void	create_default();
		int		get_mtx(uint16 no,uint16 *width,int16 *lsb);
		int		set_mtx(uint16 no,uint16 width,int16 lsb);
	private:
		BTextControl	*tc_st,*tc_ed,*tc_width,*tc_lsb;
		BSlider			*sl_left,*sl_right;
		BButton			*pb_set;
	private:
		uint16	gno;
		glyph	*g;
};

class ttf_loca:public ttf_table
{
	public:
		ttf_loca(ttf *p,uchar *dat=NULL,uint32 size=0);
		~ttf_loca();
		void	create_default();
		int		get_offset(bool f,int no,uint32 *off,uint32 *len);		//0.ok  -1.not_found
		int		chg_length(bool f,int no,uint32 len);		//0.ok  -1.not_found
	private:
};

class ttf_maxp:public ttf_table
{
	public:
		ttf_maxp(ttf *p,uchar *dat=NULL,uint32 size=0);
		~ttf_maxp();
		void	AttachedToWindow();
		void	Draw(BRect rect);
		void	MessageReceived(BMessage *msg);
		void	redraw();
		void	create_default();
		uint16	numglyphs();
		void	update();		//without numglyphs
	private:
//		tablecontrol	*tc;
//		BControl	*t;
		BTextControl	*tc1;
};

class ttf_name:public ttf_table
{
	public:
		ttf_name(ttf *p,uchar *dat=NULL,uint32 size=0);
		~ttf_name();
		void	AttachedToWindow();
		void	Draw(BRect rect);
		void	MessageReceived(BMessage *msg);
		void	redraw();
		void	create_default();
		int		count();
		int		get_field(int no,uint16 *pid,uint16 *eid,uint16 *lid,uint16 *nid,int32 *len,char *str);
		int		set_field(uint16 pid,uint16 eid,uint16 lid,uint16 nid,const char *str=NULL);
	private:
		BListView		*lv;
		BTextControl	*tc1,*tc2,*tc3,*tc4,*tc5;
		BButton			*pb1,*pb2,*pb3;
};

class ttf_post:public ttf_table
{
	public:
		ttf_post(ttf *p,uchar *dat=NULL,uint32 size=0);
		~ttf_post();
		void	AttachedToWindow();
		void	Draw(BRect rect);
		void	MessageReceived(BMessage *msg);
		void	redraw();
		void	create_default();
	private:
		BTextControl	*tc1;
};

class ttf_gasp:public ttf_table
{
	public:
		ttf_gasp(ttf *p,uchar *dat=NULL,uint32 size=0);
		~ttf_gasp();
		void	AttachedToWindow();
		void	Draw(BRect rect);
		void	MessageReceived(BMessage *msg);
		void	redraw();
		int		count();
		int		get_field(int no,uint16 *pt,uint16 *f);
		int		set_field(int no,uint16 pt,uint16 f=0);	//no=-1.auto,  pt=0.delete
	private:
		BListView		*lv;
		BTextControl	*tc_pt;
		BCheckBox		*cb_gf,*cb_gr;
		BButton			*pb_set,*pb_add,*pb_del;
};

class ttf_ebdt:public ttf_table
{
	public:
		ttf_ebdt(ttf *p,uchar *dat=NULL,uint32 size=0);
		~ttf_ebdt();
		int		decode(uint32 off,uint16 imgfmt,bitmap *bmp);	//0.ok  -1.not_found  +.err
		int		encode(const bitmap *bmp,uint32 off,uint16 imgfmt);	///restore dat
		ebdttoolview	*tv;
		ebdteditview	*ev;
	private:
};

class ttf_eblc:public ttf_table
{
	public:
		ttf_eblc(ttf *p,uchar *dat=NULL,uint32 size=0);
		~ttf_eblc();
		void	Draw(BRect rect);
		int		get_sizes(int *cnt,uchar *ppem);
		int		get_offset(bitmap *bmp,uint32 *off,uint16 *imgfmt);	//0.ok  -1.not_found
	private:
};

class ttf
{
	public:
		ttf();
		~ttf();
		int		init();				//default(empty)-table
		int		load(BFile &f);
		int		save(BFile &f);
		void	set_view(BView *v);		//v->AddChild(all_table)
		void	reset_view(BView *v);	//v->RemoveChild(all_table)
		int		get_tables(int *count,uint32 *name);
		BRect	bounds();
		int		get_glyph(int g_no,glyph *gl);
		int		set_glyph(int g_no,const glyph *gl);
		int		get_bitmap(bitmap *bm);
		int		set_bitmap(bitmap *bm);
		inittf	*ittf;
		uchar	*header;
		ttf_os2		*os2;
		ttf_cmap	*cmap;
		ttf_glyf	*glyf;
		ttf_head	*head;
		ttf_hhea	*hhea;
		ttf_hmtx	*hmtx;
		ttf_loca	*loca;
		ttf_maxp	*maxp;
		ttf_name	*name;
		ttf_post	*post;
		ttf_table	*etc[16];
		ttf_gasp	*gasp;
		ttf_ebdt	*ebdt;
		ttf_eblc	*eblc;
	private:
};
//========================================================
class glyftoolview:public BView
{
	public:
		glyftoolview(BRect frame,ttf_glyf *glyf);
		~glyftoolview();
		void	AttachedToWindow();
		void	Draw(BRect rect);
		void	MouseDown(BPoint pt);
		void	MessageReceived(BMessage *msg);
		void	redraw();
	private:
		BTextControl	*tc_chr,*tc_cno,*tc_gno;
		BButton			*pb_csc,*pb_set;
		BListView		*lv;
		BTextControl	*tc_sub;
		BButton			*pb_add,*pb_del;
		ttf_glyf	*parent;
		BBitmap	*bm;
		char	mode;			//draw mode
		uchar	psize;			//pensize
		float	sc;				//scale
	private:
		uint16	gno;
		uint16	chr;
		uchar	cps;
		uint16	compo[16];
		bool	emptyf,compof;
};

class glyfeditview:public BView
{
	public:
		glyfeditview(BRect frame,ttf_glyf *glyf);
		~glyfeditview();
		void	AttachedToWindow();
		void	Draw(BRect rect);
		void	KeyDown(const char *bytes,int32 numBytes);
		void	KeyUp(const char *bytes,int32 numBytes);
		void	MouseDown(BPoint pt);
		void	MouseUp(BPoint pt);
		void	MouseMoved(BPoint pt, uint32 code,const BMessage *msg);
		void	MessageReceived(BMessage *msg);
		void	chkinput();
		void	ptr_move(BPoint pt,uchar bt,input inp);
		void	btn_up(BPoint pt,uchar bt,input inp);
		void	btn_down(BPoint pt,uchar bt,input inp);
		float	getscale();
		void	chgscale(float sc);
		void	init();
		void	send2toolv(uchar f);
		void	redraw(char f=7);
		void	drawframe(BRect r);
	private:
		s_sel	sel_old,sel_now;
		char	now_mode;
		bool	mdf;		//mouse_button  t..down  f..up
		BPoint	mbp;		//mouse_before_point
		BPoint	mmp;		//mouse_move_point
		uchar	kin;		//keyin
		bool	mcf;		//copyable flag(for mouse-move)
		int		selm;		//select mode(2.move,3.edit,4.delete)
		int		selc;		//count of selected object
		float	scale;
	private:
		ttf_glyf	*parent;
		uint16	gno;
		uint16	chr;
		uchar	psize;
};

class ebdttoolview:public BView
{
	public:
		ebdttoolview(BRect frame,ttf_ebdt *ebdt);
		~ebdttoolview();
		void	AttachedToWindow();
		void	Draw(BRect rect);
		void	MouseDown(BPoint pt);
		void	MessageReceived(BMessage *msg);
		void	init();			//remake popupmenu
		void	redraw();
	private:
		BTextControl	*tc_gno,*tc_chr;//,*tc_pt;
		BMenuField		*mf;
		BButton			*pb_set;
		ttf_ebdt	*parent;
		BBitmap	*bm;
	private:
		uint16	gno,chr,pt;
};

class ebdteditview:public BView
{
	public:
		ebdteditview(BRect frame,ttf_ebdt *ebdt);
		~ebdteditview();
		void	AttachedToWindow();
		void	Draw(BRect rect);
		void	MouseDown(BPoint pt);
		void	MessageReceived(BMessage *msg);
		void	send2toolv(uchar f);
		void	redraw(char f=7);
	private:
		uint16	gno;
		uint16	chr;
		uint16	pt;
		bitmap	*bmp;
		ttf_ebdt	*parent;
		glyph	*g;
};

class initview:public BView
{
	public:
		initview(BRect frame);
		~initview();
		void	AttachedToWindow();
		void	Draw(BRect rect);
		void	MessageReceived(BMessage *msg);
		void	redraw();
		void	setchrs();
	private:
		BButton		*pb_set;
		BCheckBox	*cb[71];
};

class selectview:public BView
{
	public:
		selectview(BRect frame);
		~selectview();
		void	AttachedToWindow();
		void	Draw(BRect rect);
		void	MouseDown(BPoint pt);
		void	MessageReceived(BMessage *msg);
		void	redraw();
	private:
		BBitmap	*bm;
		int		tbls;
		uint32	table[32];
		int		no;			//select table no.
};

class fontedview:public BView
{
	public:
		fontedview(BRect frame);
		~fontedview();
		void	AttachedToWindow();
		void	Draw(BRect rect);
		void	MessageReceived(BMessage *msg);
		void	init();
		void	clr();
};

class fontedwindow:public BWindow
{
	public:
		fontedwindow(BRect frame);
		bool	QuitRequested();
		void	Zoom(BPoint rec_position,float rec_width,float rec_height);
		void	FrameResized(float width,float height);
		void	DispatchMessage(BMessage *message,BHandler *handler);
		void	settarget(BRect limit,BView *scrollview,BView *inputview,float scrw=0,float scrh=0);
		void	updatescrollbar();
		void	setupmenus();
		void	markmenu(const char *name,const char *item);
		BMenuBar	*mb;
		input	inp;
		fontedview	*v;
		selectview	*sv;
		initview	*iniv;
		BScrollBar	*sbv,*sbh;
		float	sw,sh;		//scroll area size
		BView	*scv,*iv;	//for scroll,dispatch(mouse)
};

class fonted:public BApplication
{
	public:
		fonted();
		void	RefsReceived(BMessage *msg);
		void	MessageReceived(BMessage *msg);
		void	AboutRequested();
		void	clear();
		void	start();
		void	start2();
		int		load(entry_ref &eref);
		int		save(entry_ref &dir,const char *name);
		fontedwindow	*w;
};

uchar*	malloc32(int bytes);
//========================================================
static BFilePanel	*fp;	//panel for load/save

BRect	canvas;
bool	canvasok;
static glyph	*obj=NULL;	//obj[n] >>> obj->sh[n]   objs >>> obj->cnt
static object	*obje=NULL;	//create now
static ttf		*ettf=NULL;	//ttf for edit

int		cnt_cp;											//shape_count_for_copy&paste
shape	**sh_cp;										//shape_for_copy&paste
//========================================================
static const uint16 iconpat[6][16]={
	{	//pick
		0x0000,0x0600,0x0600,0x0600,0x0600,0x0600,0x06db,0xc6db,
		0x67ff,0x37fe,0x3ffe,0x1ffc,0x0ff8,0x07f8,0x07f8,0x07f8
	},{	//free
		0x0000,0x0000,0x000c,0x0010,0x0020,0x0040,0x0040,0x0080,
		0x0080,0x0100,0x0380,0x0e40,0x1240,0x2180,0x0000,0x0000
	},{	//rect
		0x0000,0x0000,0x0000,0x0000,0x3ffc,0x2004,0x2004,0x2004,
		0x2004,0x2004,0x2004,0x3ffc,0x0000,0x0000,0x0000,0x0000
	},{	//ellipse
		0x0000,0x0000,0x0000,0x0000,0x03c0,0x0c30,0x1008,0x2004,
		0x2004,0x1008,0x0c30,0x03c0,0x0000,0x0000,0x0000,0x0000
	},{	//polygon
		0x0000,0x0000,0x0004,0x000c,0x0018,0x0028,0x0050,0x0090,
		0x0120,0x0220,0x0440,0x0840,0x1080,0x3f80,0x0000,0x0000
	},{	//round
		0x0000,0x0000,0x0000,0x0030,0x0048,0x0088,0x0110,0x0210,
		0x0420,0x0820,0x0840,0x1040,0x1080,0x0f00,0x0000,0x0000
	}
};

static const uchar	triangle[4][8]={
	{	//up
		0x00,0x00,0x18,0x3c,0x7e,0xff,0x00,0x00
	},{	//down
		0x00,0x00,0xff,0x7e,0x3c,0x18,0x00,0x00
	},{	//left
		0x20,0x30,0x38,0x3c,0x3c,0x38,0x30,0x20
	},{	//right
		0x04,0x0c,0x1c,0x3c,0x3c,0x1c,0x0c,0x04
	}
};
//========================================================
//================= reference UnicodeBlockObjects.h
static const struct{uint16 st,ed;char n[32];}uni_blk[]={
		0x0000,0x007F,"BASIC_LATIN",					//b0
		0x0080,0x00FF,"LATIN1_SUPPLEMENT",				//b1
		0x0100,0x017F,"LATIN_EXTENDED_A",				//b2
		0x0180,0x024F,"LATIN_EXTENDED_B",				//b3
		0x0250,0x02AF,"IPA_EXTENSIONS",					//b4
		0x02B0,0x02FF,"SPACING_MODIFIER_LETTERS",		//b5
		0x0300,0x036F,"COMBINING_DIACRITICAL_MARKS",	//b6
		0x0370,0x03CF,"BASIC_GREEK",					//b7
		0x03D0,0x03FF,"GREEK_SYMBOLS_AND_COPTIC",		//b8
		0x0400,0x04FF,"CYRILLIC",						//b9
		0x0530,0x058F,"ARMENIAN",						//b10
		0x0590,0x05CF,"BASIC_HEBREW",					//b11
		0x05D0,0x05FF,"HEBREW_EXTENDED",				//b12
		0x0600,0x0670,"BASIC_ARABIC",					//b13
		0x0671,0x06FF,"ARABIC_EXTENDED",				//b14
		0x0900,0x097F,"DEVANAGARI",						//b15
		0x0980,0x09FF,"BENGALI",						//b16
		0x0A00,0x0A7F,"GURMUKHI",						//b17
		0x0A80,0x0AFF,"GUJARATI",						//b18
		0x0B00,0x0B7F,"ORIYA",							//b19
		0x0B80,0x0BFF,"TAMIL",							//b20
		0x0C00,0x0C7F,"TELUGU",							//b21
		0x0C80,0x0CFF,"KANNADA",						//b22
		0x0D00,0x0D7F,"MALAYALAM",						//b23
		0x0E00,0x0E7F,"THAI",							//b24
		0x0E80,0x0EFF,"LAO",							//b25
		0x10A0,0x10CF,"BASIC_GEORGIAN",					//b26
		0x10D0,0x10FF,"GEORGIAN_EXTENDED",				//b27
		0x1100,0x11FF,"HANGUL_JAMO",					//b28
		0x1E00,0x1EFF,"LATIN_EXTENDED_ADDITIONAL",		//b29
		0x1F00,0x1FFF,"GREEK_EXTENDED",					//b30
		0x2000,0x206F,"GENERAL_PUNCTUATION",			//b31
		0x2070,0x209F,"SUPERSCRIPTS_AND_SUBSCRIPTS",	//b32
		0x20A0,0x20CF,"CURRENCY_SYMBOLS",				//b33
		0x20D0,0x20FF,"COMBINING_MARKS_FOR_SYMBOLS",	//b34
		0x2100,0x214F,"LETTERLIKE_SYMBOLS",				//b35
		0x2150,0x218F,"NUMBER_FORMS",					//b36
		0x2190,0x21FF,"ARROWS",							//b37
		0x2200,0x22FF,"MATHEMATICAL_OPERATORS",			//b38
		0x2300,0x23FF,"MISCELLANEOUS_TECHNICAL",		//b39
		0x2400,0x243F,"CONTROL_PICTURES",				//b40
		0x2440,0x245F,"OPTICAL_CHARACTER_RECOGNITION",	//b41	
		0x2460,0x24FF,"ENCLOSED_ALPHANUMERICS",			//b42
		0x2500,0x257F,"BOX_DRAWING",					//b43
		0x2580,0x259F,"BLOCK_ELEMENTS",					//b44
		0x25A0,0x25FF,"GEOMETRIC_SHAPES",				//b45
		0x2600,0x26FF,"MISCELLANEOUS_SYMBOLS",			//b46
		0x2700,0x27BF,"DINGBATS",						//b47
		0x3000,0x303F,"CJK_SYMBOLS_AND_PUNCTUATION",	//b48
		0x3040,0x309F,"HIRAGANA",						//b49
		0x30A0,0x30FF,"KATAKANA",						//b50
		0x3100,0x312F,"BOPOMOFO",						//b51
		0x3130,0x318F,"HANGUL_COMPATIBILITY_JAMO",		//b52
		0x3190,0x319F,"CJK_MISCELLANEOUS",				//b53
		0x3200,0x32FF,"ENCLOSED_CJK_LETTERS_AND_MONTHS",//b54
		0x3300,0x33FF,"CJK_COMPATIBILITY",				//b55
		0xAC00,0xD7AF,"HANGUL",							//b56
		0xD800,0xDBFF,"HIGH_SURROGATES",				//b57
		0xDC00,0xDFFF,"LOW_SURROGATES",					//b58
		0x4E00,0x9FFF,"CJK_UNIFIED_IDEOGRAPHS",			//b59
		0xE000,0xF8FF,"PRIVATE_USE_AREA",				//b60
		0xF900,0xFAFF,"CJK_COMPATIBILITY_IDEOGRAPHS",	//b61
		0xFB00,0xFB4F,"ALPHABETIC_PRESENTATION_FORMS",	//b62
		0xFB50,0xFDFF,"ARABIC_PRESENTATION_FORMS_A",	//b63
		0xFE20,0xFE2F,"COMBINING_HALF_MARKS",			//b64
		0xFE30,0xFE4F,"CJK_COMPATIBILITY_FORMS",		//b65
		0xFE50,0xFE6F,"SMALL_FORM_VARIANTS",			//b66
		0xFE70,0xFEFE,"ARABIC_PRESENTATION_FORMS_B",	//b67
		0xFF00,0xFFEF,"HALFWIDTH_AND_FULLWIDTH_FORMS",	//b68
//		0xFEFF,0xFEFF,"SPECIALS",						//b69
		0xFFF0,0xFFFF,"SPECIALS",						//b69
		0x0F00,0x0FBF,"TIBETAN",						//b70
		0,0,""											//endmark
};
